#![no_std]
#![no_main]
#![feature(type_alias_impl_trait)]
#![feature(never_type)]
#![allow(incomplete_features)]
#![feature(async_fn_in_trait)]

#[cfg(not(feature = "serial"))]
mod mouse_hid;

#[cfg(feature = "serial")]
mod usb_serial;

mod mouse_sensor;

use embassy_executor::Spawner;
use embassy_futures::join::join;
use embassy_rp::{
    peripherals::USB,
    spi::{CsPin, Instance},
};

use embassy_rp::peripherals::PIN_25;
use embassy_rp::spi::{ClkPin, MisoPin, MosiPin};
use embassy_rp::Peripheral;

// use embassy_time::{Duration, Timer};
#[cfg(not(feature = "serial"))]
use usbd_hid::descriptor::MouseReport;

#[cfg(not(feature = "serial"))]
use crate::{mouse_hid::mouse_config::MouseConfig, mouse_hid::MouseHID};

#[cfg(feature = "serial")]
use crate::usb_serial::{Serial, SerialConfig};

use crate::mouse_sensor::{MouseSensor, MouseSensorPins};

use {defmt_rtt as _, panic_probe as _};

#[embassy_executor::main]
async fn main(_spawner: Spawner) {
    let p = embassy_rp::init(Default::default());

    #[cfg(not(feature = "serial"))]
    mouse_hid(p.USB, mouse_sensor).await;

    #[cfg(feature = "serial")]
    serial_hid(
        p.USB,
        p.PIN_25,
        MouseSensorPins::new(p.SPI0, p.PIN_2, p.PIN_3, p.PIN_4, p.PIN_5),
    )
    .await;
}

#[cfg(not(feature = "serial"))]
async fn mouse_hid<'d, L, T, CSP>(usb: USB, mut mouse_sensor: MouseSensor<'d, L, T, CSP>)
where
    L: FnMut(&str) + 'd,
    T: Instance,
    CSP: CsPin<T> + 'd,
{
    let mut mouse_hid = MouseHID::new(usb, MouseConfig::new());

    let usb_future = MouseHID::run_usb().await;

    // Do stuff with the class!
    let hid_future = async {
        loop {
            if let Some((x, y)) = mouse_sensor.read_movement().await {
                let report = MouseReport {
                    buttons: 0,
                    x,
                    y,
                    wheel: 0,
                    pan: 0,
                };

                #[cfg(not(feature = "serial"))]
                mouse_hid.mouse_report(report).await;
            }
        }
    };

    join(hid_future, usb_future).await;
}

#[cfg(feature = "serial")]
async fn serial_hid<'d, T, CSP, Clk, Mosi, Miso, P, C, MI, MO, CS>(
    usb: USB,
    led: PIN_25,
    sensor_pins: MouseSensorPins<'d, T, Clk, Mosi, Miso, CSP, P, C, MI, MO, CS>,
) where
    T: Instance,

    CSP: CsPin<T> + 'd,
    Clk: ClkPin<T> + 'd,
    Mosi: MosiPin<T> + 'd,
    Miso: MisoPin<T> + 'd,

    P: Peripheral<P = T> + 'd,
    C: Peripheral<P = Clk> + 'd,
    MI: Peripheral<P = Mosi> + 'd,
    MO: Peripheral<P = Miso> + 'd,
    CS: Peripheral<P = CSP> + 'd,
{
    let (serial, usb_future) = Serial::new(usb, SerialConfig::new()).await;

    let mut mouse_sensor = MouseSensor::new(led, sensor_pins, serial).await;

    // Do stuff with the class!
    let hid_future = async {
        loop {
            if let Some((x, y)) = mouse_sensor.read_movement().await {
                // let report = MouseReport {
                //     buttons: 0,
                //     x,
                //     y,
                //     wheel: 0,
                //     pan: 0,
                // };

                mouse_sensor.sender().send_number(x, 10).await;
                mouse_sensor.sender().send_msg("\n").await;
                mouse_sensor.sender().send_number(y, 10).await;
                mouse_sensor.sender().send_msg("\n").await;
            }
        }
    };

    join(hid_future, usb_future).await;
}
