pub mod mouse_config;

use core::future::Future;

use embassy_rp::usb::Driver;
use embassy_rp::{interrupt, peripherals::USB};

use embassy_usb::class::hid::{HidWriter, ReportId, RequestHandler, State};
use embassy_usb::control::OutResponse;
use embassy_usb::{Builder, Config, UsbDevice};

use usbd_hid::descriptor::{MouseReport, SerializedDescriptor};

use mouse_config::MouseConfig;

static mut DEVICE_DESCRIPTOR_BUFFER: [u8; 256] = [0; 256];
static mut CONFIG_DESCRIPTOR_BUFFER: [u8; 256] = [0; 256];
static mut BOS_DESCRIPTOR_BUFFER: [u8; 256] = [0; 256];
static mut MSOS_DESCRIPTOR_BUFFER: [u8; 256] = [0; 256];
static mut CONTROL_BUFFER: [u8; 64] = [0; 64];
static mut STATE: Option<State<'static>> = None;
static mut USB: Option<UsbDevice<'static, Driver<'static, USB>>> = None;
static REQUEST_HANDLER: MyRequestHandler = MyRequestHandler {};

pub struct MouseHID<'d> {
    writer: HidWriter<'d, Driver<'d, USB>, 5>,
}

impl MouseHID<'static> {
    pub fn new(usb: USB, mouse_config: MouseConfig<'static>) -> MouseHID<'static> {
        // Create the driver, from the HAL.
        let irq = interrupt::take!(USBCTRL_IRQ);
        let driver = Driver::new(usb, irq);

        // Create embassy-usb Config
        let mut config = Config::new(
            mouse_config.vendor_id as u16,
            mouse_config.product_id as u16,
        );
        config.manufacturer = Some(mouse_config.manufacturer);
        config.product = Some(mouse_config.product);
        config.serial_number = Some(mouse_config.serial_number);
        config.device_class = 0;
        config.device_protocol = 2;
        config.max_power = 100;
        config.max_packet_size_0 = 64;

        unsafe {
            STATE = Some(State::new());
        }

        let mut builder = Builder::new(
            driver,
            config,
            unsafe { &mut DEVICE_DESCRIPTOR_BUFFER },
            unsafe { &mut CONFIG_DESCRIPTOR_BUFFER },
            unsafe { &mut BOS_DESCRIPTOR_BUFFER },
            unsafe { &mut MSOS_DESCRIPTOR_BUFFER },
            unsafe { &mut CONTROL_BUFFER },
        );

        // Create classes on the builder.
        let config = embassy_usb::class::hid::Config {
            report_descriptor: MouseReport::desc(),
            request_handler: Some(&REQUEST_HANDLER),
            poll_ms: 60,
            max_packet_size: 8,
        };

        let writer =
            HidWriter::<_, 5>::new(&mut builder, unsafe { STATE.as_mut().unwrap() }, config);

        // Build the builder.
        unsafe {
            USB = Some(builder.build());
        };

        MouseHID { writer }
    }

    pub async fn run_usb() -> impl Future<Output = !> {
        unsafe { USB.as_mut().unwrap().run() }
    }

    pub async fn mouse_report(&mut self, report: MouseReport) {
        match self.writer.write_serialize(&report).await {
            Ok(()) => {}
            Err(_e) => (),
        }
    }
}

struct MyRequestHandler {}

impl RequestHandler for MyRequestHandler {
    fn get_report(&self, _id: ReportId, _buf: &mut [u8]) -> Option<usize> {
        None
    }

    fn set_report(&self, _id: ReportId, _data: &[u8]) -> OutResponse {
        OutResponse::Accepted
    }

    fn set_idle_ms(&self, _id: Option<ReportId>, _dur: u32) {}

    fn get_idle_ms(&self, _id: Option<ReportId>) -> Option<u32> {
        None
    }
}
