use crate::usb_serial::Sender;

use super::paw3212db_tjdt::PAW3212DB_TJDT;

use embassy_rp::{
    gpio::{Level, Output},
    peripherals::PIN_25,
    spi::{ClkPin, CsPin, Instance, MisoPin, MosiPin},
    Peripheral,
};

use core::marker::PhantomData;

pub struct MouseSensorPins<'a, T, Clk, Mosi, Miso, CSP, P, C, MI, MO, CS>
where
    T: Instance,
    Clk: ClkPin<T> + 'a,
    Mosi: MosiPin<T> + 'a,
    Miso: MisoPin<T> + 'a,
    CSP: CsPin<T> + 'a,

    P: Peripheral<P = T> + 'a,
    C: Peripheral<P = Clk> + 'a,
    MI: Peripheral<P = Mosi> + 'a,
    MO: Peripheral<P = Miso> + 'a,
    CS: Peripheral<P = CSP> + 'a,
{
    pub spi: P,
    pub clk: C,
    pub mosi: MI,
    pub miso: MO,
    pub cs: CS,

    d: PhantomData<&'a ()>,
}

impl<'a, T, Clk, Mosi, Miso, CSP, P, C, MI, MO, CS>
    MouseSensorPins<'a, T, Clk, Mosi, Miso, CSP, P, C, MI, MO, CS>
where
    T: Instance,
    Clk: ClkPin<T> + 'a,
    Mosi: MosiPin<T> + 'a,
    Miso: MisoPin<T> + 'a,
    CSP: CsPin<T> + 'a,

    P: Peripheral<P = T> + 'a,
    C: Peripheral<P = Clk> + 'a,
    MI: Peripheral<P = Mosi> + 'a,
    MO: Peripheral<P = Miso> + 'a,
    CS: Peripheral<P = CSP> + 'a,
{
    pub fn new(spi: P, clk: C, mosi: MI, miso: MO, cs: CS) -> Self {
        Self {
            spi,
            clk,
            mosi,
            miso,
            cs,

            d: PhantomData,
        }
    }
}

pub struct MouseSensor<'d, S, T, CSP>
where
    S: Sender,
    T: Instance,
    CSP: CsPin<T> + 'd,
{
    pub spi_device: PAW3212DB_TJDT<'d, S, CSP, T>,

    led: Output<'d, PIN_25>,
}

impl<'d, S, T, CSP> MouseSensor<'d, S, T, CSP>
where
    S: Sender,
    T: Instance,
    CSP: CsPin<T> + 'd,
{
    pub async fn new<'a, Clk, Mosi, Miso, P, C, MI, MO, CS>(
        led: PIN_25,
        sensor_pins: MouseSensorPins<'d, T, Clk, Mosi, Miso, CSP, P, C, MI, MO, CS>,
        sender: S,
    ) -> MouseSensor<'d, S, T, CSP>
    where
        Clk: ClkPin<T> + 'a,
        Mosi: MosiPin<T> + 'a,
        Miso: MisoPin<T> + 'a,

        P: Peripheral<P = T> + 'a,
        C: Peripheral<P = Clk> + 'a,
        MI: Peripheral<P = Mosi> + 'a,
        MO: Peripheral<P = Miso> + 'a,
        CS: Peripheral<P = CSP> + 'a,
    {
        // These are implicitly used by the spi driver if they are in the correct mode

        let led = Output::new(led, Level::Low);
        let spi_device = PAW3212DB_TJDT::new(
            sensor_pins.spi,
            sensor_pins.clk,
            sensor_pins.mosi,
            sensor_pins.miso,
            sensor_pins.cs,
            sender,
        )
        .await;

        let mut me = Self { spi_device, led };

        me.led.set_high();

        // verify initialization
        let mut verified = false;

        for _ in 0..50 {
            if me.verify_product_id_1().await && me.verify_product_id_2().await {
                verified = true;
            }
        }

        if !verified {
            panic!();
        }

        me.led.set_low();

        me
    }

    async fn verify_product_id_1(&mut self) -> bool {
        const DEFAULT_VALUE: u8 = 0x30;

        DEFAULT_VALUE == self.spi_device.product_id_1().await
    }

    async fn verify_product_id_2(&mut self) -> bool {
        const DEFAULT_VALUE: u8 = 0x02;

        DEFAULT_VALUE == self.spi_device.product_id_2().await
    }

    async fn motion_detected(&mut self) -> (bool, bool, bool) {
        let reg = self.spi_device.motion_status().await;

        let motion = (reg & 0b10000000) != 0;
        let x_overflow = (reg & 0b00010000) != 0;
        let y_overflow = (reg & 0b00001000) != 0;

        (motion, x_overflow, y_overflow)
    }

    pub async fn read_movement(&mut self) -> Option<(i8, i8)> {
        let (motion, x_overflow, y_overflow) = self.motion_detected().await;

        if motion {
            if x_overflow || y_overflow {
                return None;
            }

            let x = self.spi_device.delta_x().await;
            let y = self.spi_device.delta_y().await;

            Some((x, y))
        } else {
            None
        }
    }

    pub fn sender(&mut self) -> &mut S {
        &mut self.spi_device.sender
    }
}
